#version 430 core
/*
* Copyright 2019 Vienna University of Technology.
* Institute of Computer Graphics and Algorithms.
* This file is part of the ECG Lab Framework and must not be redistributed.
*/

in VertexData {
	vec3 position_world;
	vec3 normal_world;
	vec2 uv;
} vert;

out vec4 color;

uniform int renderingMode;

uniform vec3 camera_world;

uniform vec3 materialCoefficients; // x = ambient, y = diffuse/metallic, z = specular/roughness
uniform float specularAlpha;
uniform sampler2D diffuseTexture;

struct DirectionalLight {
	vec3 color;
	vec3 direction;
};

struct PointLight {
	vec3 color;
	vec3 position;
	vec3 attenuation;
};

uniform DirectionalLight dirL;

const int POINT_LIGHT_COUNT = 128;
uniform PointLight pointLights[POINT_LIGHT_COUNT];
uniform uint pointLightIDs[POINT_LIGHT_COUNT + 1];

vec3 phong(vec3 n, vec3 l, vec3 v, vec3 diffuseC, float diffuseF, vec3 specularC, float specularF, float alpha, bool attenuate, vec3 attenuation) {
	float d = length(l);
	l = normalize(l);
	float att = 1.0;
	if(attenuate) {
		if(attenuation.x == 0 && attenuation.y==0 && attenuation.z==0) return vec3(0, 0, 0);
		att = 1.0f / (attenuation.x + d * attenuation.y + d * d * attenuation.z);
	}
	vec3 r = reflect(-l, n);
	return (diffuseF * diffuseC * max(0, dot(n, l)) + specularF * specularC * pow(max(0, dot(r, v)), alpha)) * att; 
}

vec3 Schlick(vec3 f0, vec3 f90, float VdH){
	return f0 + (f90 - f0) * pow(clamp((1 - abs(VdH)), 0.0, 1.0), 5.0);
}

vec3 BRDF_lambertian(vec3 f0, vec3 f90, vec3 diffuseColor, float specularWeight, float VdH){
	return (1.0 - specularWeight * Schlick(f0, f90, VdH)) * (diffuseColor / 3.141592653);
}

float V_GGX(float NdL, float NdV, float alphaRoughness){
	float GGXV = NdL * sqrt(NdV * NdV * (1.0 - alphaRoughness) + alphaRoughness);
	float GGXL = NdV * sqrt(NdL * NdL * (1.0 - alphaRoughness) + alphaRoughness);
	
	float GGX = GGXV + GGXL;
	if(GGX > 0.0){
		return 0.5 / GGX;
	}
	return 0.0;
}

float D_GGX(float NdH, float alphaRoughness){
	float f = (NdH * NdH) * (alphaRoughness - 1.0) + 1.0;
	return alphaRoughness / (3.141592653 * f * f);
}

vec3 BRDF_specular(vec3 f0, vec3 f90, float alphaRoughness, float specularWeight, float VdH, float NdL, float NdV, float NdH){
	vec3 fresnel = Schlick(f0, f90, VdH);
	float V = V_GGX(NdL, NdV, alphaRoughness);
	float D = D_GGX(NdH, alphaRoughness);
	
	return specularWeight * fresnel * V * D;
}

//https://github.com/KhronosGroup/glTF/blob/master/specification/2.0/README.md#appendix-b-brdf-implementation
//https://github.com/KhronosGroup/glTF-Sample-Viewer/blob/master/source/Renderer/shaders/brdf.glsl
vec3 pbr(vec3 n, vec3 l, vec3 v, vec3 lightColor, vec3 diffuseColor, float metallic, float roughness, bool attenuate, vec3 attenuation){
	float d = length(l);
	l = normalize(l);
	vec3 h = normalize(l + n);
	
	float att = 1.0;
	if(attenuate) {
		if(attenuation.x == 0 && attenuation.y==0 && attenuation.z==0) return vec3(0, 0, 0);
		att = 1.0f / (attenuation.x + d * attenuation.y + d * d * attenuation.z);
	}
	
	
	float NdH = clamp(dot(n, h), 0.0, 1.0);
	float NdV = clamp(dot(n, v), 0.0, 1.0);
	float NdL = clamp(dot(n, l), 0.0, 1.0);
	float VdH = clamp(dot(v, h), 0.0, 1.0);
	
	vec3 diffuse = vec3(0.0);
	vec3 specular = vec3(0.0);
	
	if(NdL > 0.0 || NdV > 0.0){
	
		vec3 black = vec3(0.0);
		vec3 dielectricSpecular = vec3(0.04);
	
		//vec3 diffuseFactor = mix(diffuseColor * (1 - dielectricSpecular.r), black, metallic);
		vec3 f0 = mix(dielectricSpecular, diffuseColor, metallic);
		float alpha = roughness * roughness;
		
		diffuse = BRDF_lambertian(f0, vec3(1.0), diffuseColor, 1.0, VdH) * NdL * lightColor * att;
		//diffuse = disney_lambertian(diffuseColor, NdL, NdV, NdH, roughness) * NdL * lightColor * att;
		specular = BRDF_specular(f0, vec3(1.0), alpha, 1.0, VdH, NdL, NdV, NdH) * NdL * lightColor * att;
	}
	
	return diffuse + specular;
	
}

float disney_GGX(float alpha, float NdH, vec3 n, vec3 h){
	float sinH = length(cross(n, h));
	
	return 1 / pow(alpha * alpha * NdH * NdH + sinH * sinH, 2);
}

vec3 disney_lambertian(vec3 diffuseColor, float NdL, float NdV, float NdH, float roughness){
	float f90 = 0.5 + 2 * NdH * NdH * roughness;
	return (diffuseColor / 3.141592653) * (1 + (f90 - 1) * pow(1 - NdL, 5)) * (1 + (f90 - 1) * pow(1 - NdV, 5));
}

//B. Burley (2012), Physically-Based Shading at Disney
vec3 disney_BRDF(vec3 n, vec3 l, vec3 v, vec3 lightColor, vec3 diffuseColor, float metallic, float roughness, bool attenuate, vec3 attenuation, float f90){
	float d = length(l);
	l = normalize(l);
	vec3 h = normalize(l + n);
	
	float att = 1.0;
	if(attenuate) {
		if(attenuation.x == 0 && attenuation.y==0 && attenuation.z==0) return vec3(0, 0, 0);
		att = 1.0f / (attenuation.x + d * attenuation.y + d * d * attenuation.z);
		
		if(att < 0.001) return vec3(0, 0, 0);
	}
	
	
	float NdH = clamp(dot(n, h), 0.0, 1.0);
	float NdV = clamp(dot(n, v), 0.0, 1.0);
	float NdL = clamp(dot(n, l), 0.0, 1.0);
	float VdH = clamp(dot(v, h), 0.0, 1.0);
	
	vec3 f0 = mix(vec3(0.04), diffuseColor, metallic);
	float alpha = roughness * roughness;
	
	vec3 diffuse = disney_lambertian(diffuseColor, NdL, NdV, NdH, roughness) * NdL * lightColor * att;
	//vec3 specular = disney_GGX(alpha, NdH, n, h) * Schlick(f0, vec3(1.0), VdH) * disney_GGX(pow(0.5 + roughness / 2, 2), NdH, n, h) * lightColor * att;
	vec3 specular = BRDF_specular(f0, vec3(1.0), alpha, f90, VdH, NdL, NdV, NdH) * NdL * lightColor * att;
	
	if(NdL > 0.0 && NdV > 0.0){
		//return diffuse + specular / (4 * NdL * NdV);
		return diffuse + specular;
	}else{
		return diffuse;
	}
}

void main() {	
	float alpha = texture(diffuseTexture, vert.uv).a;
	vec3 texColor = texture(diffuseTexture, vert.uv).rgb;
	
	if(alpha <= 0.5){
		discard;
	}
	
	vec3 n = normalize(vert.normal_world);
	vec3 v = normalize(camera_world - vert.position_world);
	
	color = vec4(texColor * materialCoefficients.x, 1); // ambient
	
	//Phong shading
	if(renderingMode == 0){
		color.rgb += phong(n, -dirL.direction, v, dirL.color * texColor, materialCoefficients.y, dirL.color, materialCoefficients.z, specularAlpha, false, vec3(0));
		for (int i = 1; i <= pointLightIDs[0]; i++) {
			uint id = pointLightIDs[i];
			color.rgb += phong(n, pointLights[id].position - vert.position_world, v, pointLights[id].color * texColor, materialCoefficients.y, pointLights[id].color, materialCoefficients.z, specularAlpha, true, pointLights[id].attenuation);
		}
	}
	
	//glTF BRDF shading
	if(renderingMode == 1){
		color.rgb += pbr(n, -dirL.direction, v, dirL.color, texColor, materialCoefficients.y, materialCoefficients.z, false, vec3(0));
			
		for (int i = 1; i <= pointLightIDs[0]; i++) {
			uint id = pointLightIDs[i];
			color.rgb += pbr(n, pointLights[id].position - vert.position_world, v, pointLights[id].color, texColor, materialCoefficients.y, materialCoefficients.z, true, pointLights[id].attenuation);
		}
	}
	
	//Disney BRDF shading
	if(renderingMode == 2){
		color.rgb += disney_BRDF(n, -dirL.direction, v, dirL.color, texColor, materialCoefficients.y, materialCoefficients.z, false, vec3(0), 0.8);
			
		for (int i = 1; i <= pointLightIDs[0]; i++) {
			uint id = pointLightIDs[i];
			color.rgb += disney_BRDF(n, pointLights[id].position - vert.position_world, v, pointLights[id].color, texColor, materialCoefficients.y, materialCoefficients.z, true, pointLights[id].attenuation, 0.8);
		}
	}
	
	//Disney BRDF with reduced fresnel (mainly for plants/grass)
	if(renderingMode == 3){
		color.rgb += disney_BRDF(n, -dirL.direction, v, dirL.color, texColor, materialCoefficients.y, materialCoefficients.z, false, vec3(0), 0.1);
			
		for (int i = 1; i <= pointLightIDs[0]; i++) {
			uint id = pointLightIDs[i];
			color.rgb += disney_BRDF(n, pointLights[id].position - vert.position_world, v, pointLights[id].color, texColor, materialCoefficients.y, materialCoefficients.z, true, pointLights[id].attenuation, 0.1);
		}
	}

	color.a = alpha;
}

